% Simulation Example (Main):
% 
% In the multi-group multicasting scenario, the transmit beamforming
% problem of minimizing the total transmitted power subject to SINR/QoS
% constraints at the receivers is solved using an efficient implementation
% of the iterative SOCP method (see our EUSIPCO'12 paper). The simulation
% is performed over a range of SINR thresholds and averaged over Monte
% Carlo runs.
%
% Author: Nils Bornhorst


clear all;
close all;


%% Some Settings:

% no. of multicast groups:
noGroups = 3; %3

% no. of antennas:
noAntennas = 6; %6

% no. of receivers:
noReceivers = 16; %16

% noise variance at receivers:
receiveNoisePower = 0.1; %0.1

% vector assigning users to multicast groups:
if noGroups == 3 && noReceivers == 16
    % just for the current particular setting:
    groupAssignments = [ones(5, 1); 2*ones(5, 1); 3*ones(6, 1)];
elseif noGroups ~= noReceivers
    % randomly generate vector allocating receivers to multicast groups in
    % case of multi-group multicasting:
    groupAssignments = randi(noGroups, noReceivers, 1);
else
    % use this in case of multi-user downlink beamforming:
    groupAssignments = (1 : noGroups).';
end

% no. of Monte Carlo runs:
noMonteCarlos = 30;
    
% initialisation of x-axis (SINR threshold) and y-axis (minimized total
% transmit power):
SINRthresh_dB = (-5:5)';
P_T_dB = zeros(length(SINRthresh_dB), 1);

% initialization of feasibility counts:
feasCount = zeros(length(SINRthresh_dB), 1);


%% Loop through Monte Carlo runs:

for mc = 1 : noMonteCarlos
    
    % generate vectors of channel coefficients with -
    % perfect CSI, Rayleigh fading, unit variance:
    hs = 1/(sqrt(2))*(randn(noAntennas, noReceivers) + 1i*randn(noAntennas, noReceivers));

    % use them to generate rank-one/instantaneous channel covariance
    % matrices:
    Hs = zeros(noAntennas, noAntennas, noReceivers);
    for m = 1 : noReceivers                
        % calculate  mth rank-one channel covariance matrix:
        Hs(:,:,m) = hs(:,m)*hs(:,m)';
    end
    
    
    %% Varying SINR threshold (QoS threshold):

    for s = 1 : length(SINRthresh_dB)
        % SINR threshold in linear scale:
        SINRthresh = 10^(SINRthresh_dB(s)/10);

        [w, P_T] = iterSOCPefficient(hs, SINRthresh, ...
            receiveNoisePower, noGroups, groupAssignments, 10);
        
        % solve the power control problem to check feasibility of w:
        if ~isnan(w(1)) % avoid if problem was infeasible
            [w, P_T] = solvePowerControl(w, Hs, SINRthresh, receiveNoisePower, ...
                noGroups, groupAssignments);
        end
        
        % calculate power in dB:
        if ~isnan(w(1))
            P_T_dB(s) = P_T_dB(s) + 10*log10(P_T);
        end

        % count feasible cases:
        feasCount(s) = feasCount(s) + ~isnan(w(1));
    end
    
    %% End varying SINR threshold

end % loop through Monte Carlo runs


%% Plot results:

% transmitted power vs. SINR threshold:    
figure;
plot(SINRthresh_dB, P_T_dB./feasCount, 'kx-', 'MarkerSize', 7);
xlabel('SINR threshold [dB]'); ylabel('Total transmitted power [dB]');
legend('Proposed efficient iterative SOCP method', 'Location', 'SouthEast');

% percentage of feasible Monte Carlo runs vs. SINR threshold:    
figure;
plot(SINRthresh_dB, feasCount*(100/noMonteCarlos), 'kx-', 'MarkerSize', 7);
xlabel('SINR threshold [dB]'); ylabel('Percentage of feasible Monte Carlo runs');
legend('Proposed efficient iterative SOCP method', 'Location', 'SouthWest');
