function [w, P_T] = iterSOCPefficient(hs, SINRthresh, receiveNoisePower, ...
    noGroups, groupAssignments, noIterations)

% [w, P_T] = iterSOCPefficient(hs, SINRthresh, receiveNoisePower, noGroups,
% groupAssignments, noIterations)
%
% This function solves the QoS-constrained power minimization problem in
% multi-group multicast beamforming using an efficient implementation of
% the iterative SOCP method. This efficient implementation is a combination
% of the original iterative SOCP method and the barrier method. The barrier
% method is chosen as a straight-forward example of an interior-point
% method.
%
% === Input: ===
% hs:                   Matrix whose columns are the frequency-falt channel
%                       vectors. The mth channel vector contains the
%                       complex channel coefficients from the transmit
%                       antenna array to the mth user.
% SINRthresh:           SINR threshold, minimum SINR that is guaranteed by
%                       the QoS constraints.
% receiveNoisePower:    Noise power at the receivers/users.
% noGroups:             Number of multicasting groups.
% groupAssignments:     Vector assigning the users to different multicast
%                       groups. The mth element is the index of the
%                       multicast group user m belongs to.
% noIterations:         Maximum number of iterations performed.
%
% === Output: ===
% w:                    Vector containing the beamforming weight vectors
%                       obtained by this method of all multicast groups
%                       stacked on top of each other.
% P_T:                  Total transmitted power corresponding to w.
%
% Author: Nils Bornhorst


% number of transmit antennas:
noAntennas = length(hs(:, 1));

% number of receivers:
noReceivers = length(hs(1, :));


%% New definition for conversion into SOCP (matrix for computing interference and noise):

Hs = zeros(noAntennas*noGroups + 1, noAntennas*noGroups + 1, noReceivers);
for m = 1 : noReceivers
    H_m = zeros(noAntennas*noGroups + 1, noAntennas*noGroups + 1);
    
    for k = 1 : noGroups
        if k == groupAssignments(m)
            % put a zeros block for the desired group
            H_m(((k - 1)*noAntennas + 2) : (k*noAntennas + 1), ((k - 1)*noAntennas + 2) : ...
                (k*noAntennas + 1)) = zeros(noAntennas, noAntennas);
        else
            % covariance matrix of current channel vector as block diagonals in H_m:
            H_m(((k - 1)*noAntennas + 2) : (k*noAntennas + 1), ((k - 1)*noAntennas + 2) : ...
                (k*noAntennas + 1)) = hs(:,m)*hs(:,m)';
        end
    end
    
    H_m(1, 1) = receiveNoisePower;
    
    Hs(:,:,m) = H_m^(1/2);
end


%% Some settings

% factor by which the barrier accuracy parameter t is increased (in
% iterations with insignificant rotations):
mu = 8;

% variable for counting number of iterations in phase I method to avoid
% infinite loop:
iterCount = 0;

% temporary feasibility vector z; its mth element later indicates by how
% much the mth QoS constraint is violated:
z = 10*ones(noReceivers, 1);

% initial starting w:
w = [1; 0.1*ones(noAntennas*noGroups, 1) + 1i*0.1*ones(noAntennas*noGroups, 1)];

% initialize vector storing rotations:
rotations = zeros(2*noIterations, noReceivers);

% for demonstraiting convergence of total transmit power:
P_T_converge = zeros(noIterations + 1, 1);


%% Phase I method (feasibility search):

% set initial barrier accuracy parameter:
t = 1;

while (max(z) > 0.1) && (iterCount < noIterations)
    
    % solve the logarithmic barrier approximation of the SOCP using
    % Newton's method:
    [z, w] = newtonsMethodFeas(w, hs, Hs, SINRthresh, receiveNoisePower, noGroups, ...
        groupAssignments, t);
    
    % rotate all inner products w_k'h_m onto positive real axis via
    % rotation of all hs:
    for m = 1 : noReceivers

        % extract mth h vector:
        h_m = hs(:, m);

        % extract kth w (sub-)vector where mth receiver is in kth multicast group:
        w_k = w((groupAssignments(m) - 1)*noAntennas + 2 : ...
            groupAssignments(m)*noAntennas + 1);

        rotation = angle(w_k'*h_m);
        h_m = h_m*exp(-1i*rotation);
        hs(:, m) = h_m;
        
        % calulate all necessary rotations for the decision whether to
        % increase t or not:
        rotations(iterCount + 1, m) = abs(rotation)*180/pi;

    end
    
    z;
    
    % increase barrier accuracy parameter if rotations in current iteration
    % are insignificant:
    if max(rotations(iterCount + 1, :)) < 30
        t = mu*t;
    end
    
    iterCount = iterCount + 1;
    
end % while (max(z) > 0.1) && (iterCount < noIterations)

% Calculate a strictly feasible starting point for Newton's method. The
% above feasibility search may not be sufficient for this purpose. Its
% purpose is rather to find a feasible initial approximation of the
% feasible set. Once such a feasible initial approximation has been found,
% it is better to use the basic phase I method (minimize the maximum
% infeasibility rather than the sum of feasibilities) because the single
% feasibility indicator z can be < 0:
w = basicPhaseImethod(w, hs, Hs, SINRthresh, receiveNoisePower, noGroups, ...
    groupAssignments, 5);

% calculate total transmit power for final feasibility iteration (to
% demonstrate convergence):
P_T_converge(1) = 10*log10(abs(w(2:length(w))'*w(2:length(w))));


%% Phase II method (power minimization):

% reset barrier accuracy parameter (because optimization problem may change
% significantly now):
t = 0.1;

for it = 1 : noIterations
    
    % solve the logarithmic barrier approximation of the SOCP using
    % Newton's method:
    w = newtonsMethod(w, hs, Hs, SINRthresh, receiveNoisePower, noGroups, ...
        groupAssignments, t);
   
    % only if problem was feasible, prevents hs from becoming NaN
    if ~isnan(w(1))
        
        % rotate all inner products w_k'h_m onto positive real axis via
        % rotation of all hs:
        for m = 1 : noReceivers

            % extract mth h vector:
            h_m = hs(:, m);

            % extract kth w (sub-)vector where m is in kth multicast group:
            w_k = w((groupAssignments(m) - 1)*noAntennas + 2 : ...
                groupAssignments(m)*noAntennas + 1);
            
            rotation = angle(w_k'*h_m);
            h_m = h_m*exp(-1i*rotation);
            hs(:, m) = h_m;

            % calulate all necessary rotations for the decision whether to
            % increase t or not:
            rotations(noIterations + it, m) = abs(rotation)*180/pi;
            
        end
        
    end
    
    % increase barrier accuracy parameter if rotations in current iteration
    % are insignificant:
    if max(rotations(noIterations + it, :)) < 30
        t = mu*t;
    end
    
    % calculate total transmit power for current iteration (to demonstrate
    % convergence):
    P_T_converge(it + 1) = 10*log10(abs(w(2:length(w))'*w(2:length(w))));
    
    % if we are not in the first iteration, if t was increased in the
    % previous iteration and if the change in power is negligible from the
    % previous to the current iteration -> break to avoid diverging
    % Newton's method due to an increase of t without significant update of
    % w:
    if it ~= 1 && max(rotations(noIterations + it - 1, :)) < 30 && ...
            abs(P_T_converge(it + 1) - P_T_converge(it)) < 0.3
        break;
    end
    
end


%% Results:

% optimized weight vector (remove first element which is = 1):
w = w(2:length(w));

% minimized total transmitted power:
P_T = real(w'*w); % real avoids tiny imaginary part

