function [w, P_T] = solvePowerControl(w, Hs, SINRthresh, receiveNoisePower, ...
    noGroups, groupAssignments)

% [w, P_T] = solvePowerControl(w, Hs, SINRthresh, receiveNoisePower,
% noGroups, groupAssignments)
%
% This function solves the power control problem such that the input
% beamforming weight vectors contained in w are optimally scaled, i.e., the
% total transmitted power is minimized while all QoS constraints are
% satisfied.
%
% === Input: ===
% w:                    Vector containing the beamforming weight vectors to
%                       be scaled of all multicast groups stacked on top of
%                       each other.
% Hs:                   Array containing all rank-one channel covariance
%                       matrices.
% SINRthresh:           SINR threshold, minimum SINR that is guaranteed by
%                       the QoS constraints.
% receiveNoisePower:    Noise power at the receivers/users.
% noGroups:             Number of multicasting groups.
% groupAssignments:     Vector assigning the users to different multicast
%                       groups. The mth element is the index of the
%                       multicast group user m belongs to.
%
% === Output: ===
% w:                    Vector containing the scaled beamforming weight
%                       vectors of all multicast groups.
% P_T:                  Total transmitted power corresponding to w.
%
% Author: Nils Bornhorst


% number of transmit antennas:
noAntennas = length(Hs(:, 1, 1));

% number of receivers:
noReceivers = length(Hs(1, 1, :));


%% Calculate constants for multigroup multicasting power control problem:

% initialize matrix containing receive signal powers:
Alpha = zeros(noReceivers + 1, noGroups);

% initialize vector containing transmit powers:
beta = zeros(noGroups, 1);

for m = 1 : noReceivers    
    for k = 1 : noGroups        
        % extract kth w (sub-)vector
        w_k = w((k - 1)*noAntennas + 1 : k*noAntennas);
        
        % signal power at user m from stream k:
        Alpha(m, k) = real(w_k'*Hs(:,:,m)*w_k); % real avoids tiny imaginary part
        
        if groupAssignments(m) == k
            % minus sign for desired signal (since originally on RHS of <=)
            Alpha(m, k) = -Alpha(m, k);
        else
            % times SINRthresh for interference
            Alpha(m, k) = SINRthresh*Alpha(m, k);
        end
        
        % priliminary transmit power for group k:
        beta(k) = w_k'*w_k;      
    end    
end

% last row of Alpha corresponds to the constraints p(k) >= 0 (power boost
% (or reduction) factors have to be non-negative):
Alpha(noReceivers + 1, :) = -ones(1, noGroups);


%% Solve the linear power control problem:

% isntead of cvx which is rather slow for LPs, use linprog
% (linprog(f, A, b) solves min f'*x s.t. A*x <= b):

% silence linprog:
options = optimset('Display', 'off');

% solve the power control problem:
%
% min_p sum_{k=1}^G beta(k)*p(k)
% s.t.  Alpha(m,k)/(sum_{l~=k} Alpha(m,l)*p(l) + receiveNoisePower)
%       >= SINRthresh, for all receivers in group k, for all groups
%       p >= 0
%
% p: power boost or reduction factors
[p, P_T, exitFlag] = linprog(beta, Alpha, ...
    [-SINRthresh*receiveNoisePower*ones(noReceivers, 1); 0], [],[],[],[],[], options);

% if the problem is infeasible:
if exitFlag < 1
    p = NaN*ones(noGroups, 1);
end

P_T = beta'*p;

% finally scale the beamforming vectors:
for k = 1 : noGroups    
    % scale kth w (sub-)vector
    w((k - 1)*noAntennas + 1 : k*noAntennas) = sqrt(p(k))*w((k - 1)*noAntennas + 1 : ...
        k*noAntennas);    
end

